/* $Id: ast_datatype.h 7194 2024-01-08 19:29:54Z aron $ */
/*
    Welltype parser and AST
    Copyright (C) 2014-2023  Aron Barath

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#ifndef WTPARSER_AST_DATATYPE_H
#define WTPARSER_AST_DATATYPE_H

#include "../astnode.h"
#include "welltype/wtruntime.h"
#include "welltype/wtb3.h"
#include "ast_type.h"
#include <deque>
#include <map>
#include <string>

namespace wtast
{

class DataTypeCons;

class DataType : public Type
{
	typedef std::deque<DataTypeCons*> conses_t;
	typedef std::map<std::string, DataTypeCons*> consmap_t;
	conses_t      mConses;
	consmap_t     mConsMap;
	std::string   mName;
	wt_u32        mTypeId;
	wt_u32        mNextConsId;
	DataTypeCons* mDefaultCons;

public:
	DataType(const std::string & name, wt_u32 type_id)
		: Type()
		, mName(name)
		, mTypeId(type_id)
		, mNextConsId(1)
		, mDefaultCons(0)
	{
		// ok
	}

	~DataType();

	void Accept(AstVisitor* vis);

	DataType* Copy();

	Kind GetKind() const
	{
		return DATATYPE;
	}

	std::string GetName() const
	{
		return mName;
	}

	wt_u32 GetTypeId() const
	{
		return mTypeId;
	}

	wt_u32 GetLoadSize() const;

	bool IsObject() const
	{
		return true;
	}

	bool IsDefaultConstructible() const
	{
		return 0!=mDefaultCons;
	}

	bool HasFunctionType() const;

	bool IsCopyable() const;

	void AddCons(DataTypeCons* cons);

	std::size_t GetNumberOfConses() const
	{
		return mConses.size();
	}

	DataTypeCons* GetCons(std::size_t idx)
	{
		if(idx<mConses.size())
		{
			return mConses[idx];
		}
		else
		{
			return 0;
		}
	}

	DataTypeCons* GetCons(const std::string & consname)
	{
		consmap_t::iterator it = mConsMap.find(consname);

		if(mConsMap.end()!=it)
		{
			return it->second;
		}

		return 0;
	}

	DataTypeCons* GetDefaultCons()
	{
		return mDefaultCons;
	}

	void SetDefaultCons(DataTypeCons* cons)
	{
		mDefaultCons = cons;
	}

	void SetMutable()
	{
		mTypeId |= WTB3TF_MUTABLE;
	}
};

} // namespace wtast

#endif // WTPARSER_AST_DATATYPE_H
