/* $Id: ast_variabledeclaration.cpp 6310 2023-01-16 20:39:48Z aron $ */
/*
    Welltype parser and AST
    Copyright (C) 2014-2023  Aron Barath

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "ast_variabledeclaration.h"
#include "ast_variabledeclfragment.h"
#include "ast_type.h"
#include "ast_expression.h"
#include "../astvisitor.h"

wtast::VariableDeclaration::~VariableDeclaration()
{
	delete mInitialValue;

	while(!mFragments.empty())
	{
		delete mFragments.front();
		mFragments.pop_front();
	}
}

void wtast::VariableDeclaration::Accept(AstVisitor* vis)
{
	if(vis->Visit(this))
	{
		for(fragments_t::iterator it=mFragments.begin(),end=mFragments.end();it!=end;++it)
		{
			(*it)->Accept(vis);
		}

		if(mInitialValue)
		{
			mInitialValue->Accept(vis);
		}

		vis->EndVisit(this);
	}
}

wtast::VariableDeclaration* wtast::VariableDeclaration::Copy()
{
	VariableDeclaration* vardecl = new VariableDeclaration(mType);

	for(fragments_t::iterator it=mFragments.begin(),end=mFragments.end();it!=end;++it)
	{
		vardecl->AddFragment(new VariableDeclFragment(vardecl, (*it)->GetName(), (*it)->GetUniqueId()));
	}

	if(mInitialValue)
	{
		vardecl->mInitialValue = mInitialValue->Copy();
	}

	return vardecl;
}
