#ifndef WTRUNTIME_PROGRAM_H
#define WTRUNTIME_PROGRAM_H

#include "config.h"
#include "types.h"
#include "vector.h"
#include <stddef.h>

struct wtr_vm;
struct wtr_program;
struct wtr_symtab;

struct wtr_program
{
	struct wtr_node        link_programs;         /* link in the program list of the VM */
	struct wtr_vm*         vm;                    /* the program belongs to this virtual machine */
	wt_uint64_t            prog_id;               /* unique program ID */
	wt_uintptr             refcount;              /* number of references to this program (due to other programs, or function handles) -- starts at 1 (self-reference) */
	wt_vmptr               program_data;          /* data allocated for the program (that the binary specified to the loader) */
	wt_uintptr             program_size;          /* size of 'program_data' in octets */
	void*                  user_ptr;              /* an additional user pointer, this is opaque to the VM */
	wt_u32                 pc_base;               /* program-local PC base address relative to program.code */
	wt_u32                 pc_ctor;               /* PC of the ctor{} (program-relative address; WTR_NULLPC if not exists */
	wt_u32                 pc_dtor;               /* PC of the dtor{} (program-relative address; WTR_NULLPC if not exists */
	wt_u32                 pc_main;               /* entry point of the program (in practice PC of main{}), program-relative address */
	wt_vmptr               base_address;          /* VM address of 'program_data' */
	struct wtr_symtab*     symtab;
	struct wtr_vector      referenced_symtabs;
	struct wtr_vector      metadata_vec;
};

/* metadata structure, key string and value string are allocated in a */
/* single memory block, thus freeing a metadata is achive by calling */
/* wtr_vm_sys_free(vm, metadata_ptr) */
struct wtr_metadata
{
	const char* key;
	const char* val;
};

void wtr_program_destroy(struct wtr_vm* const vm,
	struct wtr_program* const program);

void wtr_program_unref_symtabs(struct wtr_program* const prog);
void wtr_program_free_metadata(struct wtr_program* const prog);

#endif /* WTRUNTIME_PROGRAM_H */
