#ifndef WTRUNTIME_PUBLIC_H
#define WTRUNTIME_PUBLIC_H

#include "config.h"
#include <stddef.h>

#ifndef WTAPIENTRY
#  define WTAPIENTRY
#endif /* WTAPIENTRY */


/*----------------------------------------------------------------*/
/* version numbers */

#define WT_INSTRVER_BEHEMOTH    1
#define WT_INSTRVER_BEHEMOTH2   2
#define WT_INSTRVER_TOMB        3

#define WT_MAKEVERSION(_major_,_minor_,_patch_) \
	( \
		(((wt_u32)(_major_)) << 24) | \
		(((wt_u32)(_minor_)) << 16) | \
		(((wt_u32)(_patch_)) <<  8) \
	)

#define WT_GETMAJOR(_version_)  (((_version_)>>24) & 0xff)
#define WT_GETMINOR(_version_)  (((_version_)>>16) & 0xff)
#define WT_GETPATCH(_version_)  (((_version_)>> 8) & 0xff)


/*----------------------------------------------------------------*/
/* limits */

#define WTR_MAX_SYMNAME_LEN     63 /* excluding terminating zero */
#define WTR_MAX_HASH_DIGEST_LEN (512/8)


/*----------------------------------------------------------------*/
/* virtual machine flags */

#define WTRVMF_NOFLAGS         0x00000000ul /* placeholder */
#define WTRVMF_REQINTEGRITY    0x00000001ul /* loader requires validated binaries (at least one INTEGRITY section must pass) */


/*----------------------------------------------------------------*/
/* additional flags for loading programs */

#define WTRLF_NOFLAGS          0x00000000ul /* placeholder */
#define WTRLF_REQINTEGRITY     0x00000001ul /* loader requires validated binaries (at least one INTEGRITY section must pass) */


/*----------------------------------------------------------------*/
/* types IDs, and type flags */

#define WTRTF_REF               0x00000001ul
#define WTRTF_MUTABLE           0x00000002ul /* function argument type modifier: able to modify object content */
#define WTRTF_SEQ               0x00000004ul
#define WTRTF_USER              0x00000008ul
#define WTRT_TYPEID(tinfo)      ((tinfo) >> 4) /* extract type ID from typeinfo */
#define WTRT_TYPEFLAGS(tinfo)   ((tinfo) & 0xF) /* extract type flags from typeinfo */
#define WTRT_TYPEINFO(flg,id)   (((id)<<4) | ((flg) & 0xF)) /* compile typeinfo */
#define WTRT_RES_TYPEBITS       4 /* number of reserved bits in typeinfo */

#define WTTYPE_BOOL             1
#define WTTYPE_CHAR             2
#define WTTYPE_I8               3
#define WTTYPE_U8               4
#define WTTYPE_I16              5
#define WTTYPE_U16              6
#define WTTYPE_I32              7
#define WTTYPE_U32              8
#define WTTYPE_I64              9
#define WTTYPE_U64              10
#define WTTYPE_F32              11
#define WTTYPE_F64              12
#define WTTYPE_STRING           13
#define WTTYPE_HANDLE           14
#define WTTYPE_FILE             15
#define WTTYPE_DIR              16
#define WTTYPE_XSTRING          64 /* external string (API-only type, same as WTTYPE_STRING, the only difference is memory access) */

#define WTT_BOOL                WTRT_TYPEINFO(0, WTTYPE_BOOL)
#define WTT_CHAR                WTRT_TYPEINFO(0, WTTYPE_CHAR)
#define WTT_I8                  WTRT_TYPEINFO(0, WTTYPE_I8)
#define WTT_U8                  WTRT_TYPEINFO(0, WTTYPE_U8)
#define WTT_I16                 WTRT_TYPEINFO(0, WTTYPE_I16)
#define WTT_U16                 WTRT_TYPEINFO(0, WTTYPE_U16)
#define WTT_I32                 WTRT_TYPEINFO(0, WTTYPE_I32)
#define WTT_U32                 WTRT_TYPEINFO(0, WTTYPE_U32)
#define WTT_I64                 WTRT_TYPEINFO(0, WTTYPE_I64)
#define WTT_U64                 WTRT_TYPEINFO(0, WTTYPE_U64)
#define WTT_F32                 WTRT_TYPEINFO(0, WTTYPE_F32)
#define WTT_F64                 WTRT_TYPEINFO(0, WTTYPE_F64)
#define WTT_STRING              WTRT_TYPEINFO(0, WTTYPE_STRING)
#define WTT_HANDLE              WTRT_TYPEINFO(0, WTTYPE_HANDLE)
#define WTT_FILE                WTRT_TYPEINFO(0, WTTYPE_FILE)
#define WTT_DIR                 WTRT_TYPEINFO(0, WTTYPE_DIR)
#define WTT_XSTRING             WTRT_TYPEINFO(0, WTTYPE_XSTRING)


/*----------------------------------------------------------------*/
/* boolean values */

#define WT_TRUE                 1
#define WT_FALSE                0


/*----------------------------------------------------------------*/
/* byte-order macros */

#define WTBO_LITTLE             1 /* little endian */
#define WTBO_BIG                2 /* big endian */


/*----------------------------------------------------------------*/
/* special program counter value */

#define WTR_NULLPC              0


/*----------------------------------------------------------------*/
/* error codes */

#define WTE_OK                 0      /* successful */
#define WTE_BADARG             0x0001 /* bad argument */
#define WTE_FAULT              0x0002 /* bad address */
#define WTE_NOENT              0x0003 /* no such entity */
#define WTE_NOMEM              0x0004 /* not enough memory */
#define WTE_NOFIT              0x0005 /* data does not fit in the given buffer */
#define WTE_DUPENT             0x0006 /* entity duplication detected */
#define WTE_REFER              0x0007 /* reference counter is non-zero */
#define WTE_BINERR             0x0008 /* general binary error */
#define WTE_BADVER             0x0009 /* bad binary version or unsupported feature */
#define WTE_UNRES              0x0010 /* unresolved external symbol */
#define WTE_AMBIG              0x0011 /* ambiguous symbol */
#define WTE_BADEX              0x0012 /* bad external (e.g. multiple or bad operator) */
#define WTE_URTEX              0x0013 /* unhandled runtime exception */
#define WTE_INVOKE             0x0014 /* failed to invoke main{} block or function */
#define WTE_INTEG              0x0015 /* integrity check failed */
#define WTE_MEMSZ              0x0016 /* specified memory is too small */
#define WTE_REFCTX             0x0017 /* number of context references is non-zero */
#define WTE_BADCTX             0x0018 /* context in bad state */
#define WTE_NOSTK              0x0019 /* available stack is too small */
#define WTE_DESTRP             0x0020 /* program destructor invoked */
#define WTE_UCLEAN             0x0021 /* structure needs cleaning */
#define WTE_BADMAG             0x0022 /* bad magic number */
#define WTE_ABIVER             0x0023 /* binary ABI version out of supported range */
#define WTE_STKSZ              0x0024 /* stack size is invalid */
#define WTE_VINVAL             0x0025 /* invalid argument to VMMU */
#define WTE_VNSEGS             0x0026 /* invalid number of VMMU segments */
#define WTE_VSEGSZ             0x0027 /* bad VMMU segment size */
#define WTE_VINIT              0x0028 /* VMMU init failed */
#define WTE_BADHEAP            0x0029 /* bad heap configuration */
#define WTE_SYMNAME            0x0030 /* symbol name is too long */
#define WTE_HAINIT             0x0031 /* hash algorithm init failed */
#define WTE_HADIGL             0x0032 /* hash algorithm digest length out of range */
#define WTE_TYPEINF            0x0033 /* invalid type info in binary */
#define WTE_GIDFULL            0x0034 /* GID-tree is full */
#define WTE_BADREL             0x0035 /* invalid relocation type */
#define WTE_BADSYM             0x0036 /* invalid symbol data in binary */
#define WTE_NREF               0x0037 /* too many references */


/*----------------------------------------------------------------*/
/* metadata keys */

#define WT_KEY_AUTHOR          ((const char*)"Author")
#define WT_KEY_TITLE           ((const char*)"Title")
#define WT_KEY_VERSION         ((const char*)"Version")
#define WT_KEY_COPYRIGHT       ((const char*)"Copyright")
#define WT_KEY_ID              ((const char*)"Id")
#define WT_KEY_COMMENT         ((const char*)"Comment")


/*----------------------------------------------------------------*/
/* directory entry types */

#define WTDT_DIR               'D'
#define WTDT_REG               'R'
#define WTDT_BLK               'B'
#define WTDT_CHR               'C'
#define WTDT_LNK               'L'
#define WTDT_FIFO              'F'
#define WTDT_SOCK              'S'
#define WTDT_UNKNOWN           '?' /* entry type could not be determined */


/*----------------------------------------------------------------*/
/* file access modes */

#define WTFA_EXIST             0
#define WTFA_READ              4
#define WTFA_WRITE             2
#define WTFA_EXECUTE           1


/*----------------------------------------------------------------*/
/* file seek whence values */

#define WTSW_SET               1
#define WTSW_CUR               2
#define WTSW_END               3


/*----------------------------------------------------------------*/
/* types */

typedef wt_uint32_t wt_error;     /* holds an error code (WTE_*) */
typedef wt_uint32_t wt_bool;      /* WT_TRUE or WT_FALSE */
typedef wt_uint32_t wt_char;
typedef wt_uint32_t wt_globalid;
typedef wt_globalid wt_exception;
typedef wt_globalid wt_typeid;
typedef wt_uint64_t wt_handle;


/*----------------------------------------------------------------*/
/* runtime types */


/* Type of the "pointer" used in the virtual machine. This is a virtual */
/* address within the specified VMMU32. Tomb uses 32-bit addresses, which */
/* limits the maximal addressable memory to around 4 GiB. */
typedef wt_u32 wt_vmptr;

#ifndef WT_DISABLE_PRAGMA_PACK
/* TODO */
#  pragma pack(1)
#endif /* WT_DISABLE_PRAGMA_PACK */
struct wt_pfn
{
	union
	{
	void*  ptr;
	wt_u64 _pad;
	};
	wt_u32 ext;
};
#ifndef WT_DISABLE_PRAGMA_PACK
/* TODO */
#  pragma pack()
#endif /* WT_DISABLE_PRAGMA_PACK */

struct vmmu32_segment;
struct wtr_vm;
struct wtr_program;
struct wtr_context;
struct wtr_extcallinfo;
struct wtr_sym_func;

typedef struct wtr_program*  wt_program;


/*----------------------------------------------------------------*/
/* declarations for the VM */

typedef void* wt_file;
typedef void* wt_dir;

struct _wt_value
{
	union
	{
	const wt_u32 type;
	wt_u32       writable_type;
	void*        _pad;
	};
	union
	{
	wt_bool      v_bool;
	struct
	{
	wt_u32   n;
	wt_vmptr p;
	}            v_string;
	struct
	{
	wt_u32 n;
	char*  p;
	}            v_xstring;
	wt_i8        v_i8;
	wt_u8        v_u8;
	wt_i16       v_i16;
	wt_u16       v_u16;
	wt_i32       v_i32;
	wt_u32       v_u32;
	wt_i64       v_i64;
	wt_u64       v_u64;
#if WTRUNTIME_ENABLE_FLOAT32
	wt_f32       v_f32;
#endif /* WTRUNTIME_ENABLE_FLOAT32 */
#if WTRUNTIME_ENABLE_FLOAT64
	wt_f64       v_f64;
#endif /* WTRUNTIME_ENABLE_FLOAT64 */
	wt_char      v_char;
	wt_handle    v_handle;
	wt_file      v_file;
	wt_dir       v_dir;
	struct
	{
	wt_u32   n;
	wt_vmptr p;
	}            v_seq;
	};
};
typedef struct _wt_value wt_value;

struct wtr_function
{
	struct wtr_sym_func* func;
	wt_globalid          func_id;
};

/* External function. Implementation shall call exactly one of the */
/* following functions to complete the call: */
/* - wtrContextPushReturnValues(), */
/* - wtrContextRaiseException(); */
/* or: not calling any of the function above to yield from the external */
/* function and cause WTRR_EXTCALL to be returned from wtrContextRun() */
/* until the call is completed. */
typedef void (*WT_PFN_EXTFUNC)(struct wtr_context* ctx,
	struct wtr_extcallinfo* callinfo, wt_value* argv);

/* realloc()-style function with context. If 'memory' is NULL, then should */
/* allocate a memory block with the specified size; if 'size' is zero, then */
/* should deallocate the memory block; if both zero (NULL and zero), then */
/* nothing should be changed; if 'memory' is not NULL and 'size' is not */
/* zero, then realloc can happen (can return NULL if not supported). */
typedef void* (*WT_PFN_ALLOC)(void* ctx, void* memory, size_t size);

struct wtr_loaderops
{
	void (*check_failed)(void* ctx, unsigned int line);
	void (*unres_exc)(void* ctx, const char* name);
	void (*unres_func)(void* ctx, const char* name, wt_uint32_t argc,
		wt_uint32_t retc, const wt_typeid* argv, const wt_typeid* retv);
	void (*unres_type)(void* ctx, const char* name);
	void (*unres_op)(void* ctx, const char* name);
};

enum wtr_runresult
{
	/* initial state, the context is not yet set up to execute anything; */
	/* use wtrContextInvokeProgram() or wtrContextInvokeFunction() */
	WTRR_INIT = 0,

	/* there are instructions to be executed */
	WTRR_RUN,

	/* called function returned, pop return values with */
	/* wtrContextPopReturnValues(), even if number of returned values is */
	/* zero -- this will reset the context to initial state */
	WTRR_RETURNED,

	/* function consumed supplied quantum, subsequent run possible */
	WTRR_QUANTUM,

	/* internal state, just call wtrContextRun() -- the next action is to */
	/* perform to raise the scheduled exception */
	WTRR_RAISE,

	/* called function returned with an unhandled exception -- take the */
	/* exception ID with wtrContextGetException() */
	WTRR_EXCEPTION,

	/* execution suspended due to a standard I/O call -- gather info with */
	/* wtrContextGetIOCallParams(), and push return values using */
	/* wtrContextPushReturnValues() even if there are no actual return */
	/* values; or raise an exception with wtrContextRaiseException() */
	WTRR_IOCALL,

	/* execution suspended due to an external function call -- gather info */
	/* with wtrContextGetExtCallParams(), and push return values using */
	/* wtrContextPushReturnValues() even if there are no actual return */
	/* values; raise an exception with wtrContextRaiseException(), or */
	/* yield by not calling any of the function mentioned above [in this */
	/* case wtrContextRun() will return, and the external call can be */
	/* handled at the callers side] */
	WTRR_EXTCALL,
};

/* Number of arguments the I/O call with the most arguments has. */
#define WTR_IOCALL_ARGMAX       3

enum wtr_iocall
{
	/* write to standard output */
	/* (string buf) -> () */
	/* (xstring buf) -> () */
	/* Host program shall write the specified string to the standard output */
	/* by the specified length. Should not be treated as a C-string. Note */
	/* that, this I/O call might get a 'string' or an 'xstring'. */
	WTIOC_WRITESTDOUT,

	/* read a line from standard input */
	/* (xstring buf, char separator) -> (u32 len) */
	/* Host program shall read a line from the standard input, which must */
	/* meet the following conditions: number of read characters must not be */
	/* more than size of 'buf' and read shall stop when 'separator'  */
	/* character is read (and this character must not be part of the line). */
	/* If line does not fit in the supplied buffer 'IOException' shall be */
	/* raised. If the input is empty (and no characters were read), host */
	/* must raise an 'EndOfFileException'. Return value is number of */
	/* characters filled into the buffer. */
	WTIOC_READLNSTDIN,

	/* check permissions for a file */
	/* (string pathname, u32 mode) -> (bool) */
	/* Mode is OR'd combination of WTFA_* constants. */
	WTIOC_ACCESS,

	/* get file status */
	/* (string) -> (bool, char, u64, u64, u64) */
	/* Parameter is the path name of the file. Return values are: */
	/* - bool result: true if status successfully acquired, false otherwise; */
	/* - char type: type of the file, WTDT_... constants; */
	/* - u64 size: size of the file; */
	/* - u64 mtime: time of last modification (in msecs since epoch); */
	/* - u64 birthtime: time of birth (in msecs since epoch). */
	WTIOC_STAT,

	/* get file status */
	/* (file) -> (bool, char, u64, u64, u64) */
	/* Same as WTIOC_STAT, but works on open files. */
	WTIOC_FSTAT,

	/* open a file */
	/* (string pathname, char mode) -> (file) */
	/* Mode is one of 'r', 'R', 'w', 'W', 'a', 'A'. The host program shall */
	/* raise IOException if the file cannot be opened. */
	WTIOC_FOPEN,

	/* close a file */
	/* (file) -> () */
	/* Close the file, if it was opened. */
	WTIOC_FCLOSE,

	/* reposition a stream */
	/* (file, i64 offset, u32 whence) -> (i64 result) */
	/* Value in 'whence' is one of the WTSW_... macros (in case of an */
	/* invalid value, IOException should be raised). Seeking in the file */
	/* should clear end-of-file indicator. Return value should be the */
	/* resultant file offset. */
	WTIOC_FSEEK,

	/* check end-of-file indicator */
	/* (file) -> (bool) */
	/* Host should return true if EOF is reached, false otherwise. Seeking */
	/* in the file should clear end-of-file indicator. */
	WTIOC_FEOF,

	/* write to file */
	/* (file, string buf) -> (u32 len) */
	/* (file, xstring buf) -> (u32 len) */
	/* Write the specified buffer into the file. Buffer shall not be */
	/* treated as a C-string. Note that, this I/O call might get a 'string' */
	/* or an 'xstring'. */
	WTIOC_FWRITE,

	/* read from file */
	/* (file, string buf) -> (u32 len) */
	/* Read the specified amount of data into the buffer. Return value is */
	/* the actually read number of bytes; short-read is allowed. */
	WTIOC_FREAD,

	/* read a line from file */
	/* (file, xstring buf, char separator) -> (u32 len) */
	/* Same as WTIOC_READLNSTDIN, but works on opened files. */
	WTIOC_FREADLN,

	/* open a directory */
	/* (string pathname) -> (dir) */
	/* The host program shall raise 'IOException' if the directory cannot */
	/* be opened. */
	WTIOC_OPENDIR,

	/* close a directory */
	/* (dir) -> () */
	/* Close the directory, if it was opened. (If it wasn't opened, then */
	/* no operation.) */
	WTIOC_CLOSEDIR,

	/* read a directory */
	/* (dir) -> (bool) */
	/* Read the next entry in the directory. Return value shall be true if */
	/* there was a read entry; false otherwise. */
	WTIOC_READDIR,

	/* get type of the current directory entry */
	/* (dir) -> (char) */
	/* Return value must be a WTDT_... constant which indicates the type */
	/* of the current directory entry. Raising IOException is allowed if no */
	/* entry was read. */
	WTIOC_DIRGETTYPE,

	/* get the current directory entry name */
	/* (dir, xstring buf) -> (u32) */
	/* Fill the name of the current directory entry into 'buf'. If no entry */
	/* was previously read, or name is longer than size of 'buf', an */
	/* IOException shall be raised. Return the name length. */
	WTIOC_DIRGETNAME,
};

struct wtr_extcallinfo
{
	struct wtr_vm* vm;    /* the virtual machine */
	WT_PFN_EXTFUNC func;  /* pointer to the external function */
	void*          token; /* user supplied function token */
	wt_u32         argc;  /* number of arguments */
	wt_u32         retc;  /* number of return values */
	const wt_u32*  retv;  /* return value type array */
};

struct wtr_hashalg
{
	const char* name;
	wt_u16      hashalg_id;
	wt_u16      digest_len;

	/* Initialize the hash algorithm. Return value must be zero if init was */
	/* successful; non-zero otherwise. NOTE: Only 1 init() will be called */
	/* simultaneously, and it will be closed with call to final(). This */
	/* makes possible to implement algorithms without dynamic memory */
	/* management, if needed. */
	int (*init)(const struct wtr_hashalg* const alg, void** const ctx);

	void (*update)(const struct wtr_hashalg* const alg, void* const ctx,
		const void* data, size_t size);

	/* Finalize the hash, shall fill 'digest'; free up 'ctx'. */
	void (*final)(const struct wtr_hashalg* const alg, void* const ctx,
		void* digest);
};


/*----------------------------------------------------------------*/
#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */
/*----------------------------------------------------------------*/


/* Get the Welltype runtime version. Version number is a hexadecimal */
/* number made with WT_MAKEVERSION(). Returned version is comparable */
/* to WTR_VERSION. */
/* Returns: version number */
WTAPIENTRY wt_u32 wtrGetVersion(void);


/* Convert error code to its mnemonic (i.e. WTE_OK -> "OK"). */
/* Returns: string or NULL */
/*   err - error code */
WTAPIENTRY const char* wtrMnError(wt_error err);


/* Convert error code to string (i.e. WTE_OK -> "Success"). */
/* Returns: string or NULL */
/*   err - error code */
WTAPIENTRY const char* wtrStrError(wt_error err);


/* Create a new virtual machine that can execute Tomb code. */
/* Returns: error code (WTE_*) */
/*   p_vm       - pointer to the virtual machine will be returned here */
/*   flags      - OR'd WTRVMF_* constants */
/*   vma_segs   - VMMU segment array (first segment must be empty); */
/*                warning: heap segments and any acquired segment must not */
/*                be altered! */
/*   vma_nsegs  - number of VMMU segments */
/*   vma_szsh   - VMMU segment size shift (size = 1 << vma_szsh) */
/*   heap_start - virtual address where the heap starts */
/*   heap_size  - size of the heap, must cover only addressable memory */
/*   heap_gran  - heap granuality (size of the minimal allocatable memory */
/*                block); must be power of two; limits are forced */
/*   ratio_css  - CallStackSize ratio for wtrContextCreate(); actual size */
/*                is computed with: */
/*                (stack_size*ratio_css)/(ratio_css+ratio_dss), */
/*                and will be rounded towards CSS */
/*   ratio_dss  - DataStackSize ratio for wtrContextCreate(); actual size */
/*                is computed the same way as CSS */
/*   alloc_ctx  - the context which will be passed to 'alloc_func' */
/*   alloc_func - memory allocation function */
WTAPIENTRY wt_error wtrCreateTombVM(struct wtr_vm** const p_vm,
	wt_u32 flags, struct vmmu32_segment* const vma_segs, size_t vma_nsegs,
	wt_u32 vma_szsh, wt_u32 heap_start, wt_u32 heap_size, wt_u32 heap_gran,
	wt_u8 ratio_css, wt_u8 ratio_dss, void* alloc_ctx,
	WT_PFN_ALLOC alloc_func);


/* Destroy the virtual machine context. */
/* Returns: error code (WTE_*) */
/*   vm - virtual machine to destroy */
WTAPIENTRY wt_error wtrVMDestroy(struct wtr_vm* const vm);


/* Seed the VM's internal pseudo-random generator */
/* Returns: error code (WTE_*) */
WTAPIENTRY wt_error wtrVMSeedRNG(struct wtr_vm* const vm, wt_uint32_t seed);


/* Set operations used by the loader when an issue is found. */
/* Returns: error code (WTE_*) */
WTAPIENTRY wt_error wtrVMSetLoaderOps(struct wtr_vm* const vm,
	void* const ops_ctx, const struct wtr_loaderops* const ops);


/* Set hash algoritm array. Array is terminated by name==NULL. */
/* It is possible to override the hash algoritm array. Clears the array */
/* by setting it to NULL. Note: the algorithm array is not copied, only the */
/* pointer to the first element is stored. */
/* Returns: error code (WTE_*) */
WTAPIENTRY wt_error wtrVMSetHashAlgs(struct wtr_vm* const vm,
	const struct wtr_hashalg* const algs);


/* Set the opaque user pointer stored in the virtual machine. */
/* Returns: error code (WTE_*) */
/*   vm  - virtual machine to change */
/*   ptr - the new value of the per-VM user pointer */
WTAPIENTRY wt_error wtrVMSetUserPointer(struct wtr_vm* const vm, void* ptr);


/* Get the opaque user pointer stored in the virtual machine. */
/* Returns: error code (WTE_*) */
/*   vm   - virtual machine to use */
/*   pptr - store the value of the per-VM user pointer into */
WTAPIENTRY wt_error wtrVMGetUserPointer(struct wtr_vm* const vm, void** pptr);


/* Copy memory from the VM's virtual space to the specified buffer. */
WTAPIENTRY wt_error wtrVMReadMemory(struct wtr_vm* const vm, wt_vmptr vma,
	void* ptr, wt_u32 size);


/* Copy memory from the specified buffer to the VM's virtual space. */
WTAPIENTRY wt_error wtrVMWriteMemory(struct wtr_vm* const vm, wt_vmptr vma,
	const void* ptr, wt_u32 size);


/* Get the global ID of the exception. */
/* Returns: error code (WTE_*) */
/*   vm    - search context */
/*   name  - name of the exception (zero terminated string) */
/*   p_exc - write the global ID into this field */
WTAPIENTRY wt_error wtrVMGetException(struct wtr_vm* const vm,
	const char* const name, wt_exception* const p_exc);


/* Retrieve exception name from global exception ID. */
/* Returns: error code (WTE_*) */
/*   vm      - search context */
/*   exc     - global exception ID */
/*   namenuf - buffer for name */
/*   bufsize - size of the buffer in characters */
WTAPIENTRY wt_error wtrVMExceptionToString(struct wtr_vm* const vm,
	wt_exception exc, char* namebuf, size_t bufsize);


/* Add a new exception into the context. */
/* Returns: error code (WTE_*) */
/*   vm      - context */
/*   name    - name of the exception (zero terminated string) */
/*   out_exc - memory slot to store the exception ID of the new exception */
/*             (optional) */
WTAPIENTRY wt_error wtrVMAddException(struct wtr_vm* const vm,
	const char* const name, wt_exception* const out_exc);


/* Add a new function into the context. */
/* Returns: error code (WTE_*) */
/*   vm    - context */
/*   name  - name of the function (zero terminated string) */
/*   argc  - number of arguments */
/*   argv  - array of argument types (may NULL when 'argc' is zero) */
/*   retc  - number of return values */
/*   retv  - array of return value types (may NULL when 'retc' is zero) */
/*   func  - C function which will be executed when invoke the function */
/*   token - a custom token to be registered with the function */
WTAPIENTRY wt_error wtrVMAddFunction(struct wtr_vm* const vm,
	const char* name, wt_uint32_t argc, const wt_typeid* argv,
	wt_uint32_t retc, const wt_typeid* retv, WT_PFN_EXTFUNC func,
	void* token);


/* Find and get a handle to a function. */
/* Returns: error code (WTE_*) */
/*   vm     - context */
/*   p_func - write the function handle into this field */
/*   name   - name of the function (zero terminated string) */
/*   argc   - number of arguments */
/*   argv   - array of argument types (may NULL when 'argc' is zero) */
/*   retc   - number of return values */
/*   retv   - array of return value types (may NULL when 'retc' is zero) */
/*   pure   - function is pure function or not */
WTAPIENTRY wt_error wtrVMGetFunction(struct wtr_vm* const vm,
	struct wtr_function* const func, const char* name, wt_uint32_t argc,
	const wt_typeid* argv, wt_uint32_t retc, const wt_typeid* retv,
	wt_bool pure);


/* Release a function handle which returned by wtrVMGetFunction(). */
/* Returns: error code (WTE_*) */
/*   vm   - virtual machine */
/*   func - function handle to release */
WTAPIENTRY wt_error wtrVMReleaseFunction(struct wtr_vm* const vm,
	struct wtr_function* const func);


/* Load a program into the specified virtual machine. */
/* Returns: error code (WTE_*) */
/*   vm      - virtual machine */
/*   program - write the program handle into this field */
/*   ctx     - context in initial state, used to execute the optional */
/*             program constructor */
/*   data    - WTB data */
/*   size    - size of 'data' in octets */
/*   flags   - additional flags (WTRLF_*) */
WTAPIENTRY wt_error wtrVMLoadProgram(struct wtr_vm* const vm,
	wt_program* program, struct wtr_context* const ctx, const void* data,
	size_t size, wt_u32 flags);


/* Same as wtrVMLoadProgram(), but read program from VM memory. */
WTAPIENTRY wt_error wtrVMLoadProgramVM(struct wtr_vm* const vm,
	wt_program* program, struct wtr_context* const ctx, wt_vmptr data,
	size_t size, wt_u32 flags);


/* Release a program. */
/* Returns: error code (WTE_*) */
/*   vm      - virtual machine */
/*   program - program to release */
/*   ctx     - context to set up if cleanup is required to unload the program */
WTAPIENTRY wt_error wtrVMReleaseProgram(struct wtr_vm* const vm,
	wt_program program, struct wtr_context* const ctx);


/* Set the opaque user pointer stored in the program. */
/* Returns: error code (WTE_*) */
/*   vm      - virtual machine */
/*   program - program to change */
/*   ptr     - the new value of the per-program user pointer */
WTAPIENTRY wt_error wtrProgramSetUserPointer(wt_program program, void* ptr);


/* Get the opaque user pointer stored in the program. */
/* Returns: error code (WTE_*) */
/*   vm      - virtual machine */
/*   program - program to use */
/*   pptr    - store the value of the per-program user pointer into */
WTAPIENTRY wt_error wtrProgramGetUserPointer(wt_program program, void** pptr);


/* Get the value of the specified metadata. If the metadata key is not */
/* found, NULL is returned. */
WTAPIENTRY const char* wtrProgramGetMetadata(wt_program program,
	const char* const key);


/* Create a new context that can run programs in the specified virtual */
/* machine. Context can be destroyed with wtrContextDestroy(), when it is */
/* in initial state. Newly created context will start in initial state. */
/* In initial state the context can be setup to execute a function or a */
/* special entry of a program. After the set up function returned, the */
/* context can/must be reset to initial state by calling */
/* wtrContextPopReturnValues(). */
/* Returns: error code (WTE_*) */
/*   p_context  - memory slot for the pointer to the created context */
/*   vm         - the virtual machine */
/*   stack_addr - address of the pre-allocated stack (stack will be */
/*                allocated if address is 0) */
/*   stack_size - the context will have this amount of stack space (note */
/*                that, Tomb VM has separate data and call stack, and this */
/*                size will be split according to the ratio set in the VM); */
/*                minimal (128) and maximal (16MiB) stack size is enforced */
/*   sbuf_size  - string buffer to use in I/O calls (set to 0 to use the */
/*                default sbuf size; there is an internal minimal value */
/*                which will be enforced) */
/*   flags      - not used yet, must be zero */
WTAPIENTRY wt_error wtrContextCreate(struct wtr_context** const p_context,
	struct wtr_vm* const vm, wt_vmptr stack_addr, wt_u32 stack_size,
	wt_u32 sbuf_size, wt_u32 flags);


/* Destroy a context that is in initial state. When WTE_UCLEAN is returned, */
/* wtrContextCleanup() must be called, and the context may be destroyed */
/* after context returned to idle. */
/* Returns: error code (WTE_*) */
WTAPIENTRY wt_error wtrContextDestroy(struct wtr_context* const context);


/* Start a cleanup in the context. After a successful call to this function */
/* wtrContextRun() must be called, until context does not return to idle. */
/* Returns: error code (WTE_*) */
WTAPIENTRY wt_error wtrContextCleanup(struct wtr_context* const context);


/* Set the opaque user pointer stored in the context. This pointer is set */
/* to NULL when context is created. */
/* Returns: error code (WTE_*) */
/*   ptr - the new value of the per-context user pointer */
WTAPIENTRY wt_error wtrContextSetUserPointer(
	struct wtr_context* const context, void* ptr);


/* Get the opaque user pointer stored in the context. */
/* Returns: error code (WTE_*) */
/*   pptr - store the value of the per-program user pointer into */
WTAPIENTRY wt_error wtrContextGetUserPointer(
	struct wtr_context* const context, void** pptr);


/* Invoke the main{} of the specified program. Context must be in initial */
/* state. Sets up the context to executable state, see wtrContextRun(). */
/* Returns: error code (WTE_*) */
WTAPIENTRY wt_error wtrContextInvokeProgram(struct wtr_context* const context,
	wt_program program);


/* Invoke the specified function. Context must be in initial state. Number */
/* and types of the arguments must match with the functions arguments. Sets */
/* up the context to executable state, see wtrContextRun(). 'argv' can be */
/* NULL, iff 'argc' is zero. */
/* Returns: error code (WTE_*) */
WTAPIENTRY wt_error wtrContextInvokeFunction(
	struct wtr_context* const context, struct wtr_function* const func,
	wt_uint32_t argc, wt_value* argv);


/* Pop return values of the invoked entity if context state is RETURNED. */
/* Note that, some functions and main{} does not have return values, but */
/* this function must be called. Change the context state from RETURNED to */
/* INIT. Important: in case of returned object (i.e. string or seq), the */
/* memory area belong to these objects might be destroyed when the VM */
/* continues, thus it is advised to copy them if they needed for future */
/* use. */
/* Returns: error code (WTE_*) */
WTAPIENTRY wt_error wtrContextPopReturnValues(
	struct wtr_context* const context, wt_u32* const retc, wt_u32 retmax,
	wt_value* retv);


/* Start or continue the execution of instructions in the Virtual Machine. */
/* Requires a set up context. Maximal executable instructions by a single */
/* call can be limited by 'quantum' -- if the VM executes the specified */
/* amount of intructions, it will return, thus the caller can schedule */
/* other activities. */
/* Returns: reason why the function returned (WTRR_*), further processing */
/* if the retuned value is mandatory to proper event handling */
WTAPIENTRY enum wtr_runresult wtrContextRun(struct wtr_context* const context,
	wt_u32 quantum);


/* Raise an exception in the context. Requires a set up context. After */
/* raising the exception, the context state will change to RUN, this call */
/* to wtrContextRun() is required to actually handle the exception. */
/* Acquire global exception ID by calling wtrVMAddException() or */
/* wtrVMGetException(). */
/* Returns: error code (WTE_*) */
WTAPIENTRY wt_error wtrContextRaiseException(
	struct wtr_context* const context, wt_exception exception_id);


/* Get the global ID of the unhandled exception by the context. Requires */
/* a context in EXCEPTION state. Context state will be changed from */
/* EXCEPTION to INIT. */
/* Returns: error code (WTE_*) */
WTAPIENTRY wt_error wtrContextGetException(struct wtr_context* const context,
	wt_exception* const out_exception_id);


/* Get the parameters of the external function call. This function is */
/* needed only if the context is prohibited to call external functions or */
/* the external function yielded. Context state must be EXTCALL, and this */
/* state will not changed until wtrContextPushReturnValues() or */
/* wtrContextRaiseException() is called. Note that, if an object is in */
/* the arguments, its content must be copied if required for future use, */
/* otherwise they might be destroyed then the VM continues. */
/* Returns: error code (WTE_*) */
WTAPIENTRY wt_error wtrContextGetExtCallParams(
	struct wtr_context* const context,
	struct wtr_extcallinfo** const p_callinfo, wt_value** const p_argv);


/* Get parameters of the I/O call. Requires a context in IOCALL state, */
/* and this state is preserved by this call. Context state will be changed */
/* by wtrContextPushReturnValues() or  wtrContextRaiseException(). Note */
/* that, if an object is in the arguments, its content must be copied if */
/* required for future use (however, for I/O calls this is unadvised), */
/* otherwise they might be destroyed then the VM continues. Unimplemented */
/* I/O calls should be handled by raising 'IOException'. */
/* Returns: error code (WTE_*) */
WTAPIENTRY wt_error wtrContextGetIOCallParams(
	struct wtr_context* const context, enum wtr_iocall* const iocall,
	wt_u32* const argc, wt_u32 argmax, wt_value* argv);


/* Push the return values of an external function call or an I/O call. */
/* Important: this function must be called even when the call does not have */
/* actual return values, because this function is used to change the state */
/* of the context from EXTCALL or IOCALL to RUN. Number and types of the */
/* return values must match; see documentation for details. 'retv' can be */
/* NULL, iff 'retc' is zero. */
/* Returns: error code (WTE_*) */
WTAPIENTRY wt_error wtrContextPushReturnValues(
	struct wtr_context* const context, wt_u32 retc, wt_value* retv);


/*----------------------------------------------------------------*/
#ifdef __cplusplus
}
#endif /* __cplusplus */
/*----------------------------------------------------------------*/

#endif /* WTRUNTIME_PUBLIC_H */
